@@ -1,11 +1,10 @@
-;;; ghub.el --- minuscule client library for the Github API  -*- lexical-binding: t -*-
+;;; ghub.el --- minuscule client libraries for Git forge APIs  -*- lexical-binding: t -*-
 
 ;; Copyright (C) 2016-2018  Jonas Bernoulli
 
 ;; Author: Jonas Bernoulli <jonas@bernoul.li>
 ;; Homepage: https://github.com/magit/ghub
 ;; Keywords: tools
-;; Package-Requires: ((emacs "24.4") (let-alist "1.0.5"))
 
 ;; This file is not part of GNU Emacs.
 
@@ -23,22 +22,30 @@
 
 ;;; Commentary:
 
-;; Ghub is a library that provides basic support for using the Github API
-;; from Emacs packages.  It abstracts access to API resources using only
-;; a handful of functions that are not resource-specific.
+;; Ghub provides basic support for using the APIs of various Git forges
+;; from Emacs packages.  Originally it only supported the Github REST
+;; API, but now it also supports the Github GraphQL API as well as the
+;; REST APIs of Gitlab, Gitea, Gogs and Bitbucket.
 
-;; Ghub handles the creation, storage and use of access tokens using a
-;; setup wizard to make it easier for users to get started and to reduce
-;; the support burden imposed on package maintainers.  It also comes with
-;; a comprehensive manual to address the cases when things don't just
-;; work as expected or in case you don't want to use the wizard.
+;; Ghub abstracts access to API resources using only a handful of basic
+;; functions such as `ghub-get'.  These are convenience wrappers around
+;; `ghub-request'.  Additional forge-specific wrappers like `glab-put',
+;; `gtea-put', `gogs-post' and `buck-delete' are also available.  Ghub
+;; does not provide any resource-specific functions, with the exception
+;; of `FORGE-repository-id'.
+
+;; When accessing Github, then Ghub handles the creation and storage of
+;; access tokens using a setup wizard to make it easier for users to get
+;; started.  The tokens for other forges have to be created manually.
 
 ;; Ghub is intentionally limited to only provide these two essential
 ;; features — basic request functions and guided setup — to avoid being
 ;; too opinionated, which would hinder wide adoption.  It is assumed that
 ;; wide adoption would make life easier for users and maintainers alike,
-;; because then all packages that talk to the Github API could be
-;; configured the same way.
+;; because then all packages that talk to forge APIs could be configured
+;; the same way.
+
+;; Please consult the manual (info "ghub") for more information.
 
 ;;; Code:
 
@@ -59,7 +66,8 @@
 
 ;;; Settings
 
-(defconst ghub-default-host "api.github.com")
+(defconst ghub-default-host "api.github.com"
+  "The default host that is used if `ghub.host' is not set.")
 
 (defvar ghub-github-token-scopes '(repo)
   "The Github API scopes that your private tools need.
@@ -74,12 +82,12 @@ behalf of some private tool.
 
 By default the only requested scope is `repo' because that is
 sufficient as well as required for most common uses.  This and
-other scopes are documented at https://magit.vc/goto/2e586d36.
+other scopes are documented at URL `https://magit.vc/goto/2e586d36'.
 
 If your private tools need other scopes, then you have to add
 them here *before* creating the token.  Alternatively you can
 edit the scopes of an existing token using the web interface
-at https://github.com/settings/tokens.")
+at URL `https://github.com/settings/tokens'.")
 
 (defvar ghub-override-system-name nil
   "If non-nil, the string used to identify the local machine.
@@ -93,9 +101,11 @@ used instead.")
                (:constructor ghub--make-req)
                (:copier nil))
   (url        nil :read-only nil)
+  (forge      nil :read-only t)
   (silent     nil :read-only t)
   (method     nil :read-only t)
   (headers    nil :read-only t)
+  (handler    nil :read-only t)
   (unpaginate nil :read-only nil)
   (noerror    nil :read-only t)
   (reader     nil :read-only t)
@@ -114,25 +124,7 @@ used instead.")
 (defvar ghub-response-headers nil
   "The headers returned in response to the last request.
 `ghub-request' returns the response body and stores the
-response header in this variable.")
-
-(cl-defun ghub-graphql (graphql &optional variables
-                                &key username auth host
-                                silent
-                                callback errorback extra)
-  "Make a GraphQL request using GRAPHQL and VARIABLES.
-Return the response as a json-like alist.  Even if the response
-contains `errors', do not raise an error.  GRAPHQL is a GraphQL
-string.  VARIABLES is a json-like alist.  The other arguments
-behave like for `ghub-request' (which see)."
-  (cl-assert (stringp graphql))
-  (cl-assert (not (stringp variables)))
-  (ghub-request "POST" "/graphql" nil :payload
-                (json-encode `(("query" . ,graphql)
-                               ,@(and variables `(("variables" ,@variables)))))
-                :silent silent
-                :username username :auth auth :host host
-                :callback callback :errorback errorback :extra extra))
+response headers in this variable.")
 
 (cl-defun ghub-head (resource &optional params
                               &key query payload headers
@@ -222,12 +214,12 @@ Like calling `ghub-request' (which see) with \"DELETE\" as METHOD."
                                &key query payload headers
                                silent unpaginate noerror reader
                                username auth host forge
-                               callback errorback extra)
+                               callback errorback value extra)
   "Make a request for RESOURCE and return the response body.
 
 Also place the response header in `ghub-response-headers'.
 
-METHOD is the http method, given as a string.
+METHOD is the HTTP method, given as a string.
 RESOURCE is the resource to access, given as a string beginning
   with a slash.
 
@@ -260,7 +252,7 @@ If NOERROR is non-nil, then do not raise an error if the request
   return the error payload instead of nil.
 If READER is non-nil, then it is used to read and return from the
   response buffer.  The default is `ghub--read-json-payload'.
-  For the very few resources that do not return json, you might
+  For the very few resources that do not return JSON, you might
   want to use `ghub--decode-payload'.
 
 If USERNAME is non-nil, then make a request on behalf of that
@@ -275,8 +267,8 @@ Each package that uses `ghub' should use its own token. If AUTH
   to identify itself, using a symbol matching its name.
 
   Package authors who find this inconvenient should write a
-  wrapper around this function and possibly for the method
-  specific functions also.
+  wrapper around this function and possibly for the
+  method-specific functions as well.
 
   Some symbols have a special meaning.  `none' means to make an
   unauthorized request.  `basic' means to make a password based
@@ -296,7 +288,7 @@ If HOST is non-nil, then connect to that Github instance.  This
   argument.
 
 If FORGE is `gitlab', then connect to Gitlab.com or, depending
-  on HOST to another Gitlab instance.  This is only intended for
+  on HOST, to another Gitlab instance.  This is only intended for
   internal use.  Instead of using this argument you should use
   function `glab-request' and other `glab-*' functions.
 
@@ -309,7 +301,7 @@ If CALLBACK and/or ERRORBACK is non-nil, then make one or more
 
 Both callbacks are called with four arguments.
   1. For CALLBACK, the combined value of the retrieved pages.
-     For ERRORBACk, the error that occured when retrieving the
+     For ERRORBACK, the error that occured when retrieving the
      last page.
   2. The headers of the last page as an alist.
   3. Status information provided by `url-retrieve'. Its `:error'
@@ -335,27 +327,26 @@ Both callbacks are called with four arguments.
            (error "PARAMS and PAYLOAD are mutually exclusive for METHOD %S"
                   method))
          (setq payload params)))
-  (when payload
-    (unless (stringp payload)
-      (setq payload (json-encode-list payload)))
-    (setq payload (encode-coding-string payload 'utf-8)))
   (when (or callback errorback)
     (setq noerror t))
   (ghub--retrieve
-   payload
+   (ghub--encode-payload payload)
    (ghub--make-req
     :url (url-generic-parse-url
           (concat "https://" host resource
                   (and query (concat "?" (ghub--url-encode-params query)))))
+    :forge forge
     :silent silent
     ;; Encode in case caller used (symbol-name 'GET). #35
     :method     (encode-coding-string method 'utf-8)
     :headers    (ghub--headers headers host auth username forge)
+    :handler    'ghub--handle-response
     :unpaginate unpaginate
     :noerror    noerror
     :reader     reader
     :callback   callback
     :errorback  errorback
+    :value      value
     :extra      extra)))
 
 (defun ghub-continue (req)
@@ -364,15 +355,15 @@ Both callbacks are called with four arguments.
 This function is only intended to be called from callbacks.  If
 there is a next page, then retrieve that and return the buffer
 that the result will be loaded into, or t if the process has
-already completed.  Otherwise return nil.
+already completed.  If there is no next page, then return nil.
 
 Callbacks are called with four arguments (see `ghub-request').
 The forth argument is a `ghub--req' struct, intended to be passed
-to this function.  A callbacks may use the struct's `extra' slot
+to this function.  A callback may use the struct's `extra' slot
 to pass additional information to the callback that will be
 called after the next request has finished.  Use the function
 `ghub-req-extra' to get and set the value of this slot."
-  (and (assq 'next (ghub-response-link-relations))
+  (and (assq 'next (ghub-response-link-relations req))
        (or (ghub--retrieve nil req) t)))
 
 (cl-defun ghub-wait (resource &optional duration &key username auth host)
@@ -403,16 +394,41 @@ See `ghub-request' for information about the other arguments."
                 (cl-incf total wait))
             (sit-for (setq total 2))))))))
 
-(defun ghub-response-link-relations (&optional headers)
+(defun ghub-response-link-relations (req &optional headers payload)
   "Return an alist of link relations in HEADERS.
-If optional HEADERS is nil, then return those
-in `ghub-response-headers'."
-  (let ((rels (cdr (assoc "Link" (or headers ghub-response-headers)))))
-    (and rels (mapcar (lambda (elt)
-                        (pcase-let ((`(,url ,rel) (split-string elt "; ")))
-                          (cons (intern (substring rel 5 -1))
-                                (substring url 1 -1))))
-                      (split-string rels ", ")))))
+If optional HEADERS is nil, then return those that were
+previously stored in the variable `ghub-response-headers'.
+
+When accessing a Bitbucket instance then the link relations
+are in PAYLOAD instead of HEADERS, making their API merely
+RESTish and forcing this function to append those relations
+to the value of `ghub-response-headers', for later use when
+this function is called with nil for PAYLOAD."
+  (if (eq (ghub--req-forge req) 'bitbucket)
+      (if payload
+          (let* ((page (cl-mapcan (lambda (key)
+                                    (when-let ((elt (assq key payload)))
+                                      (list elt)))
+                                  '(size page pagelen next previous)))
+                 (headers (cons (cons 'link-alist page) headers)))
+            (if (and req (or (ghub--req-callback req)
+                             (ghub--req-errorback req)))
+                (setq-local ghub-response-headers headers)
+              (setq-default ghub-response-headers headers))
+            page)
+        (cdr (assq 'link-alist ghub-response-headers)))
+  (when-let ((rels (cdr (assoc "Link" (or headers ghub-response-headers)))))
+    (mapcar (lambda (elt)
+              (pcase-let ((`(,url ,rel) (split-string elt "; ")))
+                (cons (intern (substring rel 5 -1))
+                      (substring url 1 -1))))
+            (split-string rels ", ")))))
+
+(cl-defun ghub-repository-id (owner name &key username auth host forge)
+  "Return the id of the specified repository."
+  (let ((fn (intern (format "%s-repository-id" (or forge 'ghub)))))
+    (funcall (if (eq fn 'ghub-repository-id) 'ghub--repository-id fn)
+             owner name :username username :auth auth :host host)))
 
 ;;;; Internal
 
@@ -422,11 +438,13 @@ in `ghub-response-headers'."
            (if (functionp headers) (funcall headers) headers)))
         (url-request-method (ghub--req-method req))
         (url-request-data payload)
-        (url (ghub--req-url req))
-        (silent (ghub--req-silent req)))
+        (url-show-status nil)
+        (url     (ghub--req-url req))
+        (handler (ghub--req-handler req))
+        (silent  (ghub--req-silent req)))
     (if (or (ghub--req-callback  req)
             (ghub--req-errorback req))
-        (url-retrieve url 'ghub--handle-response (list req) silent)
+        (url-retrieve url handler (list req) silent)
       ;; When this function has already been called, then it is a
       ;; no-op.  Otherwise it sets `url-registered-auth-schemes' among
       ;; other things.  If we didn't ensure that it has been run, then
@@ -438,7 +456,7 @@ in `ghub-response-headers'."
           (let ((url-registered-auth-schemes
                  '(("basic" ghub--basic-auth-errorback . 10))))
             (url-retrieve-synchronously url silent))
-        (ghub--handle-response (car url-callback-arguments) req)))))
+        (funcall handler (car url-callback-arguments) req)))))
 
 (defun ghub--handle-response (status req)
   (let ((buffer (current-buffer)))
@@ -449,14 +467,13 @@ in `ghub-response-headers'."
                  (headers    (ghub--handle-response-headers status req))
                  (payload    (ghub--handle-response-payload req))
                  (payload    (ghub--handle-response-error status payload req))
-                 (value      (nconc (ghub--req-value req) payload))
+                 (value      (ghub--handle-response-value payload req))
                  (next       (cdr (assq 'next (ghub-response-link-relations
-                                               headers)))))
+                                               req headers payload)))))
             (when (numberp unpaginate)
               (cl-decf unpaginate))
             (setf (ghub--req-url req)
                   (url-generic-parse-url next))
-            (setf (ghub--req-value req) value)
             (setf (ghub--req-unpaginate req) unpaginate)
             (or (and next
                      unpaginate
@@ -502,27 +519,38 @@ in `ghub-response-headers'."
                 payload
               (setcdr (last err) (list payload))
               nil)
-          (pcase-let ((`(,symb . ,data) err))
-            (if (eq symb 'error)
-                (if (eq (car-safe data) 'http)
-                    (signal 'ghub-http-error
-                            (let ((code (car (cdr-safe data))))
-                              (list code
-                                    (nth 2 (assq code url-http-codes))
-                                    payload)))
-                  (signal 'ghub-error data))
-              (signal symb data))))
+          (ghub--signal-error err payload))
       payload)))
 
+(defun ghub--signal-error (err &optional payload)
+  (pcase-let ((`(,symb . ,data) err))
+    (if (eq symb 'error)
+        (if (eq (car-safe data) 'http)
+            (signal 'ghub-http-error
+                    (let ((code (car (cdr-safe data))))
+                      (list code
+                            (nth 2 (assq code url-http-codes))
+                            payload)))
+          (signal 'ghub-error data))
+      (signal symb data))))
+
+(defun ghub--handle-response-value (payload req)
+  (setf (ghub--req-value req)
+        (nconc (ghub--req-value req)
+               (if-let ((nested (and (eq (ghub--req-forge req) 'bitbucket)
+                                     (assq 'values payload))))
+                   (cdr nested)
+                 payload))))
+
 (defun ghub--handle-response-payload (req)
   (funcall (or (ghub--req-reader req)
                'ghub--read-json-payload)
            url-http-response-status))
 
-(defun ghub--read-json-payload (status)
+(defun ghub--read-json-payload (_status)
   (let ((raw (ghub--decode-payload)))
     (and raw
-         (condition-case err
+         (condition-case nil
              (let ((json-object-type 'alist)
                    (json-array-type  'list)
                    (json-key-type    'symbol)
@@ -530,9 +558,15 @@ in `ghub-response-headers'."
                    (json-null        nil))
                (json-read-from-string raw))
            (json-readtable-error
-            (if (memq status (list 400 422 500))
-                "and Github didn't return JSON"
-              (signal (car err) (cdr err))))))))
+            `((message
+               . ,(if (looking-at "<!DOCTYPE html>")
+                      (if (re-search-forward
+                           "<p>\\(?:<strong>\\)?\\([^<]+\\)" nil t)
+                          (match-string 1)
+                        "error description missing")
+                    (string-trim (buffer-substring (point) (point-max)))))
+              (documentation_url
+               . "https://github.com/magit/ghub/wiki/Github-Errors")))))))
 
 (defun ghub--decode-payload (&optional _status)
   (and (not (eobp))
@@ -540,6 +574,13 @@ in `ghub-response-headers'."
         (buffer-substring-no-properties (point) (point-max))
         'utf-8)))
 
+(defun ghub--encode-payload (payload)
+  (and payload
+       (progn
+         (unless (stringp payload)
+           (setq payload (json-encode-list payload)))
+         (encode-coding-string payload 'utf-8))))
+
 (defun ghub--url-encode-params (params)
   (mapconcat (lambda (param)
                (pcase-let ((`(,key . ,val) param))
@@ -610,6 +651,20 @@ has to provide several values including their password."
                host scopes))
     scopes))
 
+;;;###autoload
+(defun ghub-clear-caches ()
+  "Clear all caches that might negatively affect Ghub.
+
+If a library that is used by Ghub caches incorrect information
+such as a mistyped password, then that can prevent Ghub from
+asking the user for the correct information again.
+
+Set `url-http-real-basic-auth-storage' to nil
+and call `auth-source-forget+'."
+  (interactive)
+  (setq url-http-real-basic-auth-storage nil)
+  (auth-source-forget+))
+
 ;;;; Internal
 
 (defun ghub--headers (headers host auth username forge)
@@ -620,22 +675,24 @@ has to provide several values including their password."
       (setq username (ghub--username host forge)))
     (lambda ()
       (if (eq auth 'basic)
-          (if (eq forge 'gitlab)
-              (error "Gitlab does not support basic authentication")
-            (cons (cons "Authorization" (ghub--basic-auth host username))
-                  headers))
+          (cons (cons "Authorization" (ghub--basic-auth host username))
+                headers)
         (cons (ghub--auth host auth username forge) headers)))))
 
 (defun ghub--auth (host auth &optional username forge)
   (unless username
     (setq username (ghub--username host)))
   (if (eq auth 'basic)
-      (if (eq forge 'gitlab)
-          (error "Gitlab does not support basic authentication")
-        (cons "Authorization" (ghub--basic-auth host username)))
-    (cons (if (eq forge 'gitlab)
-              "Private-Token"
-            "Authorization")
+      (cl-ecase forge
+        ((nil github gitea gogs bitbucket)
+         (cons "Authorization" (ghub--basic-auth host username)))
+        (gitlab
+         (error "Gitlab does not support basic authentication")))
+    (cons (cl-ecase forge
+            ((nil github gitea gogs bitbucket)
+             "Authorization")
+            (gitlab
+             "Private-Token"))
           (concat
            (and (not (eq forge 'gitlab)) "token ")
            (encode-coding-string
@@ -683,31 +740,62 @@ has to provide several values including their password."
                 ;; Auth-Source caches the information that there is no
                 ;; value, but in our case that is a situation that needs
                 ;; fixing so we want to keep trying by invalidating that
-                ;; information.  The (:max 1) is needed for Emacs releases
-                ;; before 26.1.
-                (auth-source-forget (list :max 1 :host host :user user))
+                ;; information.
+                ;; The (:max 1) is needed and has to be placed at the
+                ;; end for Emacs releases before 26.1.  See #24, #64.
+                (auth-source-forget (list :host host :user user :max 1))
                 (and (not nocreate)
-                     (if (eq forge 'gitlab)
-                         (error
-                          (concat
-                           "Required Gitlab token does not exist.  See "
-                           "https://magit.vc/manual/ghub/Gitlab-Support.html "
-                           "for instructions."))
-                       (ghub--confirm-create-token host username package)))))))
+                     (cl-ecase forge
+                       ((nil github)
+                        (ghub--confirm-create-token host username package))
+                       ((gitlab gitea gogs bitbucket)
+                        (error "Required %s token (%S for %S) does not exist.
+See https://magit.vc/manual/ghub/Support-for-Other-Forges.html for instructions."
+                               (capitalize (symbol-name forge))
+                               user host))))))))
     (if (functionp token) (funcall token) token)))
 
 (defun ghub--host (&optional forge)
-  (if (eq forge 'gitlab)
-      (or (ignore-errors (car (process-lines "git" "config" "gitlab.host")))
-          (bound-and-true-p glab-default-host))
-    (or (ignore-errors (car (process-lines "git" "config" "github.host")))
-        ghub-default-host)))
+  (cl-ecase forge
+    ((nil github)
+     (or (ignore-errors (car (process-lines "git" "config" "github.host")))
+         ghub-default-host))
+    (gitlab
+     (or (ignore-errors (car (process-lines "git" "config" "gitlab.host")))
+         (bound-and-true-p glab-default-host)))
+    (gitea
+     (or (ignore-errors (car (process-lines "git" "config" "gitea.host")))
+         (bound-and-true-p gtea-default-host)))
+    (gogs
+     (or (ignore-errors (car (process-lines "git" "config" "gogs.host")))
+         (bound-and-true-p gogs-default-host)))
+    (bitbucket
+     (or (ignore-errors (car (process-lines "git" "config" "bitbucket.host")))
+         (bound-and-true-p buck-default-host)))))
 
 (defun ghub--username (host &optional forge)
-  (let ((var (cond ((string-prefix-p "api.github.com" host) "github.user")
-                   ((string-prefix-p "gitlab.com/api" host) "gitlab.user")
-                   ((eq forge 'gitlab)     (format "gitlab.%s.user" host))
-                   (t                      (format "github.%s.user" host)))))
+  (let ((var
+         (cl-ecase forge
+           ((nil github)
+            (if (equal host ghub-default-host)
+                "github.user"
+              (format "github.%s.user" host)))
+           (gitlab
+            (if (equal host "gitlab.com/api/v4")
+                "gitlab.user"
+              (format "gitlab.%s.user" host)))
+           (bitbucket
+            (if (equal host "api.bitbucket.org/2.0")
+                "bitbucket.user"
+              (format "bitbucket.%s.user" host)))
+           (gitea
+            (when (zerop (call-process "git" nil nil nil "config" "gitea.host"))
+              (error "gitea.host is set but always ignored"))
+            (format "gitea.%s.user" host))
+           (gogs
+            (when (zerop (call-process "git" nil nil nil "config" "gogs.host"))
+              (error "gogs.host is set but always ignored"))
+            (format "gogs.%s.user"  host)))))
     (condition-case nil
         (car (process-lines "git" "config" var))
       (error
@@ -768,15 +856,36 @@ WARNING: The token will be stored unencrypted in %S.
          If you don't want that, you have to abort and customize
          the `auth-sources' option.\n" (car auth-sources))
               ""))))
-        (progn
-          (when (ghub--get-token-id host username package)
-            (if (yes-or-no-p
-                 (format
-                  "A token named %S\nalready exists on Github.  Replace it?"
-                  ident))
-                (ghub--delete-token host username package)
-              (user-error "Abort")))
-          (ghub-create-token host username package scopes))
+        (condition-case ghub--create-token-error
+            ;; Naively attempt to create the token since the user told us to
+            (ghub-create-token host username package scopes)
+          ;; The API _may_ respond with the fact that a token of the name
+          ;; we wanted already exists. At this point we're out of luck. We
+          ;; don't have a token (otherwise why would we be here?) and, if
+          ;; the user is using SMS 2FA, we have no way of telling GitHub
+          ;; to send a new 2FA code to the user other than sending a POST
+          ;; to /authorizations which is ugly.
+          ;;
+          ;; If they are not using SMS 2FA then we could try to delete the
+          ;; existing token (which will require them to hand us another
+          ;; OTP for the delete request) and then call create again,
+          ;; possibly requiring _another_ OTP if they don't do things fast
+          ;; enough, but this is only because non-SMS 2FA doesn't require
+          ;; any action on GitHub's part.
+          ;;
+          ;; GitHub does hand us a header that indicates what type of 2FA
+          ;; is in use, but it's not currently available in this location
+          ;; and would make the following code which is already quite
+          ;; complicated even more complicated. So in the interest of
+          ;; simplicity it's better to error out here and ask the user to
+          ;; take action. This situation should almost never arise anyway.
+          (ghub-http-error
+           (if (string-equal (let-alist (nth 3 ghub--create-token-error)
+                               (car .errors.code))
+                             "already_exists")
+               (error "\
+A token named %S ready exists on Github. \
+Please visit https://github.com/settings/tokens and delete it." ident))))
       (user-error "Abort"))))
 
 (defun ghub--get-token-id (host username package)
@@ -820,7 +929,8 @@ WARNING: The token will be stored unencrypted in %S.
 
 (defun ghub--auth-source-get (keys &rest spec)
   (declare (indent 1))
-  (let ((plist (car (apply #'auth-source-search :max 1 spec))))
+  (let ((plist (car (apply #'auth-source-search
+                           (append spec (list :max 1))))))
     (mapcar (lambda (k)
               (plist-get plist k))
             keys)))
@@ -836,7 +946,5 @@ line."
 
 ;;; _
 (provide 'ghub)
-;; Local Variables:
-;; indent-tabs-mode: nil
-;; End:
+(require 'ghub-graphql)
 ;;; ghub.el ends here
